import fastavro
import os
import json
import networkx as nx
import numpy as np

class NetworkLearnerObserver():
    def __init__(self, **kwargs) -> None:
        self.avro_file_name_with_path= kwargs["avro_path"]
        self.avro_file_name = ""
        self.json_file_name = ""
        self.network = None
        self.data = []
        self.all_adjacentMatrix = []
        self.latest_args = None 
        
    def update(self, args):
        self.__create_avro_file(args)

    def __create_avro_file(self, args):
        schema = args["schema_avro"]
        avro_file_name = self.avro_file_name_with_path
        #json_file_name = args["file_name"]+".json"
        
        data = args["data"]
        os.makedirs(os.path.dirname(avro_file_name), exist_ok=True)
        
        if os.path.isfile(avro_file_name):
            with open(avro_file_name, "a+b") as f_out:
                fastavro.writer(f_out, schema, data,codec="deflate")
        else:    
            with open(avro_file_name, "wb") as f_out:
                fastavro.writer(f_out, schema, data,codec="deflate")
                
        self.latest_args = args
        #self.__create_json_from_avro(avro_file_name, json_file_name)
        #self.create_network(data)
        #self.data = data
        #self.all_adjacentMatrix.append(data[0]["adjacent_matrix"])

    def __create_json_from_avro(self, avro_file_name, json_file_name):
        with open(avro_file_name, "rb") as avro_file:
            reader = fastavro.reader(avro_file)
            
            information = [info for info in reader]
        
        with open(json_file_name, "w") as json_file:
            json.dump(information, json_file, indent=4)
   
    def create_network(self, data):
        network = nx.Graph()
        adjacent_matrix = np.array(data[0]["adjacent_matrix"])
        nodes = np.arange(0, adjacent_matrix.shape[0]).tolist() 
        network.add_nodes_from(nodes)
        for fixed_node in range(adjacent_matrix.shape[0]):
            for relation_node in range(adjacent_matrix.shape[0]):
                network.add_weighted_edges_from([(fixed_node, relation_node,adjacent_matrix[fixed_node][relation_node])])
                
        self.network = network
        
    def add_AUC_to_avro_file(self, value):
        schema = self.latest_args["schema_avro"]
        avro_file_name = self.avro_file_name_with_path
        data =  self.latest_args["data"]
        data[0]['AUC_ROC'] = value
        
        if os.path.isfile(avro_file_name):
            with open(avro_file_name, "a+b") as f_out:
                fastavro.writer(f_out, schema, data,codec="deflate")
        else:    
            with open(avro_file_name, "wb") as f_out:
                fastavro.writer(f_out, schema, data,codec="deflate")
        

                     
        
        


from apache_beam.io import ReadFromAvro
import apache_beam as beam
from apache_beam.pipeline import PipelineOptions
# from apache_beam.dataframe.convert import to_dataframe
# from scripts.analysis import get_df_from_path

output = "/data/output_test/"


def main(argv=None):
    
    with beam.Pipeline() as pipeline:

        runs = (
            pipeline 
            | "read" >> ReadFromAvro(f"{output}individual*")
            | "print" >> beam.Map(print)
            # | "set schema" >> beam.Map (
            #     lambda x: beam.Row(
            # path=x['path'], 
            # model=x['model'],
            # fit_time=x['metrics'][0]['value'],
            # roc_time=x['metrics'][1]['value'],
            # roc_value=x['metrics'][2]['value']
            # )
            )
        
        # df = to_dataframe(runs)
        # df.to_csv(f"{output}individual.csv", index=False)


if __name__ == "__main__":
    main()


import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F

class CVAE(nn.Module):
    def __init__(self):
        super(CVAE, self).__init__()

        # encode
        self.fc1 = nn.Linear(33, 400)  # feature_size + class_size = 1 + 32
        self.fc21 = nn.Linear(400, 64)  # Dimensão reduzida para z_mu
        self.fc22 = nn.Linear(400, 64)  # Dimensão reduzida para z_var

        # decode
        self.fc3 = nn.Linear(64 + 32, 400)  # z (64) + c (32)
        self.fc4 = nn.Linear(400, 1)  # Reconstruir x com dimensão 1

        self.elu = nn.ELU()
        self.sigmoid = nn.Sigmoid()

    def encode(self, x, c):  # Q(z|x, c)
        '''
        x: (bs, 1)  # feature_size
        c: (bs, 32)  # class_size
        '''
        inputs = torch.cat([x, c], dim=1)  # (bs, feature_size + class_size = 33)
        h1 = self.elu(self.fc1(inputs))
        z_mu = self.fc21(h1)
        z_var = self.fc22(h1)
        return z_mu, z_var

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def decode(self, z, c):  # P(x|z, c)
        '''
        z: (bs, 64)  # latent_size
        c: (bs, 32)  # class_size
        '''
        inputs = torch.cat([z, c], dim=1)  # (bs, latent_size + class_size = 96)
        h3 = self.elu(self.fc3(inputs))
        return self.sigmoid(self.fc4(h3))

    def forward(self, x, c):
        mu, logvar = self.encode(x, c)
        z = self.reparameterize(mu, logvar)
        return self.decode(z, c), mu, logvar

    def loss_function(self, recon_x, x, mu, logvar):
        MSE = F.mse_loss(recon_x, x)
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
        return MSE + KLD



from dataclasses import dataclass


@dataclass
class ModelParam:
    name: str
    obj: object
    
@dataclass
class Params:
    path: str
    model: ModelParam

@dataclass
class Result:
    time_elapsed: float
    fun_return: object

@dataclass
class ParamResult:
    params: Params
    result: Result



"""Base classes for all mfcc estimators."""

import numpy as np
import sys
from sklearn.base import BaseEstimator, TransformerMixin
import scipy.stats as st
import librosa as lib
from functools import reduce

class Array2MelSpec(BaseEstimator, TransformerMixin):
    def __init__(self, 
                 sampling_rate,
                 n_fft,
                 hop_length,
                 n_mels,
                 frames,
                 power
                 ):
        self.sampling_rate = sampling_rate
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.n_mels = n_mels
        self.frames = frames
        self.power = power

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X, y=None, **fit_params):
        
        def normalize_melspec(mel_spectrogram):

            # 03 convert melspectrogram to log mel energy
            log_mel_spectrogram = 20.0 / self.power * np.log10(mel_spectrogram + sys.float_info.epsilon)

            # 04 calculate total vector size
            vectorarray_size = len(log_mel_spectrogram[0, :]) - self.frames + 1

            # 05 skip too short clips
            if vectorarray_size < 1:
                return np.empty((0, dims), float)

            # 06 generate feature vectors by concatenating multi_frames
            dims = self.n_mels * self.frames
            vectorarray = np.zeros((vectorarray_size, dims), float)
            for t in range(self.frames):
                vectorarray[:, self.n_mels * t: self.n_mels * (t + 1)] = log_mel_spectrogram[:, t: t + vectorarray_size].T

            return vectorarray
            
        def extract_melspec(y):
            params = {
                'y':y, 
                'n_fft': self.n_fft,
                'hop_length': self.hop_length,
                "n_mels": self.n_mels,
                "power": self.power
            }
            if self.sampling_rate:
                params['sr']=self.sampling_rate
            
            mel_spectrogram = lib.feature.melspectrogram(**params)
            
            normalized_melspec = normalize_melspec(mel_spectrogram)

            return normalized_melspec
        Xt = np.array(
            reduce(
                lambda n1, n2: np.concatenate([n1,n2]), 
                map(extract_melspec, X))
        )
        return Xt
    
    
class Array2Mfcc(BaseEstimator, TransformerMixin):
    """
     Gets a numpy array containing audio signals and transforms it into a numpy array containing mfcc signals
     
    """
    
    def __init__(self, sampling_rate):
        self.sampling_rate = sampling_rate
        #self.n_mfcc = n_mfcc

    def fit(self, X, y=None, **fit_params):
        return self

    def transform(self, X, y=None, **fit_params):
        def extract_mfcc(y):
            if self.sampling_rate: 
                return lib.feature.mfcc(y=y, sr=self.sampling_rate)
            else: 
                return lib.feature.mfcc(y=y)
        Xt = np.array(list(map(extract_mfcc, X)))
        return Xt




import os
import fastavro

class NetworkLearnerModelResultOrchestrator:
    @staticmethod
    def save_result(network, metricResult, networkName, avro_file_name):
        schema = {
            "type": "record",
            "name": "GANF_Network",
            "fields": [
                {"name": "adj_matrix_name", "type": "string"},
                {"name": "adjacent_matrix", "type": {"type": "array", "items": {"type": "array", "items": "double"}}},
                {"name": "AUC_ROC", "type": "float"},
            ],
        }

        data = [{
            "adj_matrix_name": networkName,  
            "adjacent_matrix": network,    
            "AUC_ROC": float(metricResult),  
        }]
        
        os.makedirs(os.path.dirname(avro_file_name), exist_ok=True)
        
        if os.path.isfile(avro_file_name):
            with open(avro_file_name, "a+b") as f_out:
                fastavro.writer(f_out, schema, data, codec="deflate")
        else:    
            with open(avro_file_name, "wb") as f_out:
                fastavro.writer(f_out, schema, data, codec="deflate")



import random

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from torch.nn.init import xavier_uniform_
from torch.nn.utils import clip_grad_value_
from torch.utils.data import DataLoader
from tqdm import tqdm

from mtsa.models.networkAnalysis.networkLearnerModel import NetworkLearnerModel
from mtsa.models.GANF_components.ganfLayoutData import GANFData
from mtsa.models.GANF_components.gnn import GNN
from mtsa.models.GANF_components.vae import CVAE


class GACVAEBaseModel(NetworkLearnerModel, nn.Module):
    def __init__(
        self,
        input_size=1,
        hidden_size=32,
        dropout=0.0,
        rho_max=float(1e16),
        max_iteraction=2,
        learning_rate=float(1e-3),
        alpha=0.0,
        weight_decay=float(5e-4),
        epochs=20,
        device=None,
    ):
        super().__init__()
        self.min = 0
        self.max = 0
        self.adjacent_matrix = None
        self.rho_max = rho_max
        self.max_iteraction = max_iteraction
        self.learning_rate = learning_rate
        self.alpha = alpha
        self.weight_decay = weight_decay
        self.epochs = epochs
        self.hidden_state = None

        if device != None:
            self.device = device
        else:
            self.device = torch.device(
                f"cuda:{str(str(0))}" if torch.cuda.is_available() else "cpu"
            )

        # Reducing dimensionality
        self.rnn = nn.LSTM(
            input_size=input_size,
            hidden_size=hidden_size,
            batch_first=True,
            dropout=dropout,
            device=self.device,
        )

        # Graph Neural Networks model
        self.gcn = GNN(input_size=hidden_size, hidden_size=hidden_size)

        # probability density estimation
        self.cvae = CVAE()
        self.cvae.to(device=self.device)
        self.to(device=self.device)

    @property
    def name(self):
        return "GACVAEBaseModel " + "+".join([f[0] for f in self.features])

    def get_adjacent_matrix(self):
        return self.adjacent_matrix
    
    def set_adjacent_matrix(self, adjacent_matrix):
        self.adjacent_matrix = adjacent_matrix
        
    def get_initial_adjacent_matrix(self):
        return self.initial_adjacent_matrix

    def fit(
        self,
        X,
        y=None,
        batch_size=None,
        epochs=None,
        max_iteraction=None,
        learning_rate=None,
        debug_dataframe=None,
        isWaveData=False,
        mono=None,
    ):
        torch.cuda.manual_seed(10)
        random.seed(10)
        np.random.seed(10)
        torch.manual_seed(10)
        torch.autograd.set_detect_anomaly(True)

        self.isMonoData = mono
        loss_best = 100
        self.isWaveData = isWaveData
        h_A_old = np.inf
        h_tol = float(1e-6)
        rho = 1.0

        if epochs is not None:
            self.epochs = epochs
        if max_iteraction is not None:
            self.max_iteraction = max_iteraction
        if learning_rate is not None:
            self.learning_rate = learning_rate
        if batch_size is None:
            batch_size = 32

        self.batch_size = batch_size
        dimension = X.data.shape[1]
        self.adjacent_matrix = self.__init_adjacent_matrix(X.data)
        self.initial_adjacent_matrix = self.adjacent_matrix.detach().clone()

        dataloaders = self.__create_dataLoader(
            torch.tensor(X), batch_size=batch_size, isMonoData=mono
        )
        adjacent_matrix = self.adjacent_matrix

        self.__fitCore(
            loss_best, h_A_old, h_tol, dimension, dataloaders, adjacent_matrix, rho
        )

    def predict(self, X):
        return self.__forward(X, self.adjacent_matrix)

    def score_samples(self, X):
        if self.isWaveData:
            return self.__score_wave_data(X=X)
        return self.__score_discrete_data(X=X)

    def __forward(self, x, A):
        # x: N X K X L X D
        full_shape = x.shape

        # reshape: N*K, L, D
        x = x.reshape((x.shape[0] * x.shape[1], x.shape[2], x.shape[3]))
        h, _ = self.rnn(x)

        # resahpe: N, K, L, H
        h = h.reshape((full_shape[0], full_shape[1], h.shape[1], h.shape[2]))

        h = self.gcn(h, A)

        # reshappe N*K*L,H
        h = h.reshape((-1, h.shape[3]))
        x = x.reshape((-1, full_shape[3]))

        recon_batch, mu, logvar = self.cvae(x, h)
        loss = self.cvae.loss_function(recon_batch, x, mu, logvar)

        return -loss

    def __fitCore(
        self, loss_best, h_A_old, h_tol, dimension, dataloaders, adjacent_matrix, rho
    ):
        foward_count = 0
        for j in range(self.max_iteraction):
            while rho < self.rho_max:
                learning_rate = self.learning_rate  # np.math.pow(0.1, epoch // 100)
                optimizer = torch.optim.Adam(
                    [
                        {
                            "params": self.parameters(),
                            "weight_decay": self.weight_decay,
                        },
                        {"params": [adjacent_matrix]},
                    ],
                    lr=learning_rate,
                    weight_decay=0.0,
                )
                
                for epoch in tqdm(range(self.epochs), desc=f'trainning epochs for {str(j + 1)} iteraction.', unit='epochs'):
                    loss_train = []
                    epoch += 1
                    self.train()

                    for dataloader in dataloaders:
                        for x in dataloader:
                            x = x.to(self.device)

                            optimizer.zero_grad()
                            loss = -self.__forward(x, adjacent_matrix)
                            h = (torch.trace(torch.matrix_exp(adjacent_matrix * adjacent_matrix)) - dimension)
                            total_loss = loss + 0.5 * rho * h * h + self.alpha * h

                            h.to(self.device)
                            
                            if epoch == int(.5 * self.epochs) or epoch == int(.3 * self.epochs) or foward_count == 0:
                                networkName = "first matrix" if foward_count == 0 else "intermediate matrix"
                                super().notify_observers(adjacent_matrix, epoch, h, total_loss, networkName, learning_rate=learning_rate, batch_size=self.batch_size)

                            total_loss.backward()
                            clip_grad_value_(self.parameters(), 1)
                            optimizer.step()
                            loss_train.append(loss.item())
                            adjacent_matrix.data.copy_(
                                torch.clamp(adjacent_matrix.data, min=0, max=1)
                            )
                            foward_count += 1
                            if np.mean(loss_train) < loss_best:
                                loss_best = np.mean(loss_train)
                                self.adjacent_matrix = adjacent_matrix

                del optimizer
                torch.cuda.empty_cache()

                if h.item() > 0.5 * h_A_old:
                    rho *= 10
                else:
                    break

            h_A_old = h.item()
            self.alpha += rho * h.item()

            if h_A_old <= h_tol or rho >= self.rho_max:
                break
            
            super().notify_observers(adjacent_matrix, epoch, h, total_loss, "final matrix", learning_rate=learning_rate, batch_size=self.batch_size)
            

    def __score_discrete_data(self, X):
        X = np.array(X)
        X = X.reshape(X.shape[0], X.shape[1], 1, 1)
        X = torch.tensor(X).float()
        X = X.to(self.device)
        result = self.predict(X=X)
        return float(result)

    def __score_wave_data(self, X):
        dataloaders = self.__create_dataLoader(
            X, window_size=1, isMonoData=self.isMonoData
        )
        result = []
        for dataloader in dataloaders:
            for x in dataloader:
                x = x.to(self.device)
                result.append(self.predict(X=x))
        return torch.tensor(result).mean()

    def __create_dataLoader(self, X, batch_size=32, window_size=12, isMonoData=True):
        X = self.__create_dataframe(X, isMonoData)
        dataloaders = []
        for x in X:
            x_dataloader = DataLoader(
                GANFData(x, window_size=window_size),
                batch_size=batch_size,
                shuffle=False,
                num_workers=0,
                persistent_workers=False,
            )
            dataloaders.append(x_dataloader)
        return dataloaders

    def __create_dataframe(self, X, isMonoData):
        if self.isWaveData and not isMonoData:
            dataframes = []
            for x in X:
                x = x.T
                x_df = self.__array2Dataframe(x)
                dataframes.append(x_df)
            return dataframes

        if self.isWaveData and isMonoData:
            X = X.T

        X_df = self.__array2Dataframe(X)

        return [X_df]

    def __array2Dataframe(self, x):
        x_df = pd.DataFrame(x)
        x_df = x_df.reset_index()
        x_df = x_df.rename(columns={"index": "channel"})
        x_df["channel"] = pd.to_datetime(x_df["channel"], unit="s")
        x_df = x_df.set_index("channel")
        x_df = self.__normalize(x_df)
        return x_df

    def __normalize(self, X):
        mean = X.values.mean()
        std = X.values.std()
        X = (X - mean) / std
        return X

    def __init_adjacent_matrix(self, X):
        torch.cuda.manual_seed(10)
        random.seed(10)
        np.random.seed(10)
        torch.manual_seed(10)
        init = torch.zeros([X.shape[1], X.shape[1]])
        init = xavier_uniform_(init).abs()
        init = init.fill_diagonal_(0.0)
        adjacentMatrix = torch.tensor(init, requires_grad=True, device=self.device)
        return adjacentMatrix

    def get_random_adjacent_matrix(self):
        random_state = torch.get_rng_state()
        torch.seed()
        init = torch.zeros([20, 20])
        init = xavier_uniform_(init).abs()
        init = init.fill_diagonal_(0.0)
        torch.set_rng_state(random_state)
        return torch.tensor(init, requires_grad=True, device=self.device)


from sklearn import metrics
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.pipeline import FeatureUnion, Pipeline
from sklearn.svm import OneClassSVM
from mtsa.features.mel import Array2Mfcc
from mtsa.features.stats import CrestFactor, FeatureExtractionMixer, FrequencyCenter, ImpulseValue, Kurtosis, KurtosisFactor, MarginFactor, Peak2Peak, RootFrequencyVariance, RootMeanSquareFeature, RootMeanSquareFrequency, ShapeFactor, Skewness, SquareRootOfAmplitude
from mtsa.utils import Wav2Array
from mtsa.features.stats import FEATURES
import time

class OSVM(BaseEstimator, OutlierMixin):
    def __init__(self, 
                 use_array2mfcc = False, 
                 sampling_rate=None,
                 use_featureUnion= False):
        super().__init__()
        self.sampling_rate = sampling_rate
        self.final_model = OneClassSVM(kernel="rbf", nu =0.1)
        self.use_array2mfcc = use_array2mfcc
        self.use_featureUnion = use_featureUnion
        self.model = self._build_model()
        self.last_fit_time = 0
        
    @property
    def name(self):
        return "OSVM " + "+".join([f[0] for f in self.features])
        
    def fit(self, X, y=None):    
        start = time.perf_counter()     
        self.model.fit(X)
        end = time.perf_counter()
        self.last_fit_time = end - start
        return self.last_fit_time #seconds
        
    def score_samples(self, X):
        pred = self.model.predict(X)
        pred[pred == 1] = 1
        pred[pred == -1] = 0
        pred = pred.astype(float)
        return pred
    
    def score(self, X, Y):
        X = self.score_samples
        fpr, tpr, thresholds = metrics.roc_curve(Y, X)
        auc = metrics.auc(fpr, tpr)
        return auc

    def predict(self, X):
        return self.model.predict(X)

    def _build_model(self):
        wav2array = Wav2Array(sampling_rate=self.sampling_rate)
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
        features = FeatureUnion(FEATURES)
        featureExtractionMixer = FeatureExtractionMixer()
        
        featureExtractionMixer.append_strategy(RootMeanSquareFeature())
        featureExtractionMixer.append_strategy(SquareRootOfAmplitude())
        featureExtractionMixer.append_strategy(Kurtosis())
        featureExtractionMixer.append_strategy(Skewness())
        featureExtractionMixer.append_strategy(Peak2Peak())
        featureExtractionMixer.append_strategy(CrestFactor())
        featureExtractionMixer.append_strategy(ImpulseValue())
        featureExtractionMixer.append_strategy(MarginFactor())
        featureExtractionMixer.append_strategy(ShapeFactor())
        featureExtractionMixer.append_strategy(KurtosisFactor())
        featureExtractionMixer.append_strategy(FrequencyCenter())
        featureExtractionMixer.append_strategy(RootMeanSquareFrequency())
        featureExtractionMixer.append_strategy(RootFrequencyVariance())
        
        
        if self.use_array2mfcc and self.use_featureUnion:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("array2mfcc", array2mfcc),
                    ("features", features),
                    ("final_model", self.final_model),
                ]
            )
        elif self.use_array2mfcc:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("array2mfcc", array2mfcc),
                    ("final_model", self.final_model),
                ]
            )
        elif self.use_featureUnion:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("features", features),
                    ("final_model", self.final_model),
                ]
            )
        else:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("featureExtractionMixer", featureExtractionMixer),
                    ("final_model", self.final_model),
                ]
            )

        return model



import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset

class GANFData(Dataset):
    def __init__(self, df, window_size=12, stride_size=1):
        super(GANFData, self).__init__()
        self.df = df
        self.window_size = window_size
        self.stride_size = stride_size

        self.data, self.idx, self.time = self.preprocess(df)
    
    def preprocess(self, df):
        start_idx = np.arange(0,len(df)-self.window_size,self.stride_size)
        end_idx = np.arange(self.window_size, len(df), self.stride_size)

        delat_time =  df.index[end_idx]-df.index[start_idx]
        idx_mask = delat_time==pd.Timedelta(self.window_size,unit='s')

        return df.values, start_idx[idx_mask], df.index[start_idx[idx_mask]]

    def __len__(self):

        length = len(self.idx)

        return length

    def __getitem__(self, index):
        #  N X K X L X D 
        start = self.idx[index]
        end = start + self.window_size
        data = self.data[start:end].reshape([self.window_size,-1, 1])

        return torch.FloatTensor(data).transpose(0,1)


import json
import torch
import logging
import time
import torch

import torch.optim as optim

class AETrainer:

    def __init__(self, device: str = 'cuda'):

        self.device = device

    def train(self, train_loader, ae_net, args):
        logger = logging.getLogger()

        # Set device for network
        ae_net = ae_net.to(self.device)

        # Set optimizer (Adam optimizer for now)
        optimizer = optim.Adam(ae_net.parameters(), lr=args.lr, weight_decay=args.weight_decay)

        # Training
        print('Starting pretraining...')
        start_time = time.time()
        ae_net.train()
        for epoch in range(10):


            loss_epoch = 0.0
            n_batches = 0
            epoch_start_time = time.time()
            for data in train_loader:

                if isinstance(data, list):
                    data = data[0]
                
                x = data.to(self.device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])

                # Zero the network parameter gradients
                optimizer.zero_grad()

                # Update network parameters via backpropagation: forward + backward + optimize
                outputs = ae_net(inputs)
                scores = torch.sum((outputs - inputs) ** 2, dim=tuple(range(1, outputs.dim())))
                loss = torch.mean(scores)
                loss.backward()
                optimizer.step()

                loss_epoch += loss.item()
                n_batches += 1

            # log epoch statistics
            epoch_train_time = time.time() - epoch_start_time
            print('  Epoch {}/{}\t Time: {:.3f}\t Loss: {:.8f}'
                        .format(epoch + 1, 10, epoch_train_time, loss_epoch / n_batches))

        pretrain_time = time.time() - start_time
        print('Pretraining time: %.3f' % pretrain_time)
        print('Finished pretraining.')

        return ae_net

class DeepSVDDTrainer:

    def __init__(self, device: str = 'cuda'):

        self.device = device
        # Deep SVDD parameters
       
        self.c = None


    def train(self, train_loader, net, args):
        self.args = args

        # Set device for network
        net = net.to(self.device)

        # Set optimizer (Adam optimizer for now)
        optimizer = optim.Adam(net.parameters(), lr=args.lr, weight_decay=args.weight_decay)

        # Set learning rate scheduler
        scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[20], gamma=0.1)

        # Initialize hypersphere center c (if c not loaded)
        if self.c is None:
            print('Initializing center c...')
            self.c = self.init_center_c(train_loader, net)
            print(self.c.shape)
            print('Center c initialized.')

        # Training
        print('Starting training...')
        start_time = time.time()
        net.train()

        save_path = os.path.join(args.output_dir,args.name)
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        for epoch in range(args.n_epochs):

            scheduler.step()

            loss_epoch = 0.0
            n_batches = 0
            epoch_start_time = time.time()

            for data in train_loader:
                x = data.to(self.device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])

                # Zero the network parameter gradients
                optimizer.zero_grad()

                # Update network parameters via backpropagation: forward + backward + optimize
                outputs = net(inputs).squeeze().mean(dim=-1)
                dist = torch.sum((outputs - self.c) ** 2, dim=1)

                loss = torch.mean(dist)
                loss.backward()
                optimizer.step()

                loss_epoch += loss.item()
                n_batches += 1

            # log epoch statistics
            epoch_train_time = time.time() - epoch_start_time
            print('  Epoch {}/{}\t Time: {:.3f}\t Loss: {:.8f}'
                        .format(epoch + 1, args.n_epochs, epoch_train_time, loss_epoch / n_batches))
            torch.save({'c': self.c, 'net_dict': net.state_dict()}, os.path.join(save_path, "{}_{}.pt".format(args.name, epoch)))
        self.train_time = time.time() - start_time
        print('Training time: %.3f' % self.train_time)

        print('Finished training.')

        return net

    def init_center_c(self, train_loader, net, eps=0.1):
        """Initialize hypersphere center c as the mean from an initial forward pass on the data."""
        n_samples = 0
        c = 0.0

        net.eval()
        with torch.no_grad():
            for data in train_loader:
                # get the inputs of the batch
                x = data.to(self.device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])
                outputs = net(inputs).squeeze() 
                n_samples += outputs.shape[0]
                c += torch.sum(outputs, dim=0)

        c /= n_samples

        # If c_i is too close to 0, set to +-eps. Reason: a zero unit can be trivially matched with zero weights.
        c[(abs(c) < eps) & (c < 0)] = -eps
        c[(abs(c) < eps) & (c > 0)] = eps

        return c.mean(dim=-1)


from models.RNN import RecurrentAE
from models.GAN import CNNAE


class DeepSVDD(object):

    def __init__(self, n_features, hidden_size, device):

        self.c = None  # hypersphere center c


        self.trainer = None

        # if encoder=='RNN':
        # self.ae_net = RecurrentAE(n_features, hidden_size, device)
        self.ae_net = CNNAE(n_features, hidden_size).to(device)
        self.net = self.ae_net.encoder

        self.ae_trainer = None
        self.results = {
            'test_auc': None
        }

    def train(self, dataset, args, device: str = 'cuda'):
        """Trains the Deep SVDD model on the training data."""

        self.trainer = DeepSVDDTrainer(device=device)
        # Get the model
        self.trainer.train(dataset, self.net, args)
        self.c = self.trainer.c

    def test(self, test_loader, delta_t, sigma, device):
        from utils import roc_auc_all
        import numpy as np
        self.net.eval()
        self.net.to(device)
        loss = []

        with torch.no_grad():
            for data in test_loader:
                
                x = data.to(device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])
                outputs = self.net(inputs).squeeze().mean(dim=-1)
                batch_loss= torch.sum((outputs - self.c) ** 2, dim=1).cpu().numpy()
                loss.append(batch_loss)
        loss = np.concatenate(loss)

        auc_score, fps,tps = roc_auc_all(loss, delta_t, sigma)
        print("meann: {:.4f}, median: {:.4f}, auc:{:.4f}".format(np.mean(loss), np.median(loss),auc_score))# %%
        self.results['test_auc'] = auc_score
        return auc_score, fps,tps

    def pretrain(self, train_loader, args, device):
        """Pretrains the weights for the Deep SVDD network \phi via autoencoder."""

        self.ae_trainer = AETrainer(device=device)
        self.ae_net = self.ae_trainer.train(train_loader, self.ae_net, args)
        self.net = self.ae_net.encoder

    def save_model(self, export_model):
        """Save Deep SVDD model to export_model."""

        net_dict = self.net.state_dict()

        torch.save({'c': self.c,
                    'net_dict': net_dict}, export_model)

    def load_model(self, model_path):
        """Load Deep SVDD model from model_path."""

        model_dict = torch.load(model_path)

        self.c = model_dict['c']
        self.net.load_state_dict(model_dict['net_dict'])

    def save_results(self, export_json):
        """Save results dict to a JSON-file."""
        with open(export_json, 'w') as fp:
            json.dump(self.results, fp)

# %%
import os
class DeepSADTrainer:

    def __init__(self, device: str = 'cuda'):

        self.device = device
       
        self.c = None


    def train(self, train_loader, net, args):
        self.args = args

        # Set device for network
        net = net.to(self.device)

        # Set optimizer (Adam optimizer for now)
        optimizer = optim.Adam(net.parameters(), lr=args.lr, weight_decay=args.weight_decay)

        # Set learning rate scheduler
        scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[20], gamma=0.1)

        # Initialize hypersphere center c (if c not loaded)
        if self.c is None:
            print('Initializing center c...')
            self.c = self.init_center_c(train_loader, net)
            print('Center c initialized.')

        # Training
        print('Starting training...')
        start_time = time.time()
        net.train()

        save_path = os.path.join(args.output_dir,args.name)
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        for epoch in range(args.n_epochs):



            loss_epoch = 0.0
            n_batches = 0
            epoch_start_time = time.time()

            for data, semi_targets in train_loader:

                x = data.to(self.device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])

                semi_targets = semi_targets.to(self.device)
                # Zero the network parameter gradients
                optimizer.zero_grad()

                # Update network parameters via backpropagation: forward + backward + optimize
                outputs = net(inputs).squeeze().mean(dim=-1)
                dist = torch.sum((outputs - self.c) ** 2, dim=1)

                losses = torch.where(semi_targets == 0, dist, args.eta * ((dist + 1e-6) ** semi_targets.float()))
                loss = torch.mean(losses)
                loss.backward()
                optimizer.step()

                loss_epoch += loss.item()
                n_batches += 1
            scheduler.step()
            # log epoch statistics

            epoch_train_time = time.time() - epoch_start_time
            print('  Epoch {}/{}\t Time: {:.3f}\t Loss: {:.8f}'
                        .format(epoch + 1, args.n_epochs, epoch_train_time, loss_epoch / n_batches))
            torch.save({'c': self.c, 'net_dict': net.state_dict()}, os.path.join(save_path, "{}_{}.pt".format(args.name, epoch)))

        self.train_time = time.time() - start_time
        print('Training time: %.3f' % self.train_time)

        print('Finished training.')

        return net

    def init_center_c(self, train_loader, net, eps=0.1):
        """Initialize hypersphere center c as the mean from an initial forward pass on the data."""
        n_samples = 0
        c = 0.0

        net.eval()
        with torch.no_grad():
            for data, _ in train_loader:
                # get the inputs of the batch
                x = data.to(self.device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])
                outputs = net(inputs).squeeze() 
                n_samples += outputs.shape[0]
                c += torch.sum(outputs, dim=0)

        c /= n_samples

        # If c_i is too close to 0, set to +-eps. Reason: a zero unit can be trivially matched with zero weights.
        c[(abs(c) < eps) & (c < 0)] = -eps
        c[(abs(c) < eps) & (c > 0)] = eps

        return c.mean(dim=-1)
class DeepSAD(object):

    def __init__(self, n_features, hidden_size, device):

        self.c = None  # hypersphere center c


        self.trainer = None

        self.ae_net = CNNAE(n_features, hidden_size).to(device)
        self.net = self.ae_net.encoder

        self.ae_trainer = None
        self.results = {
            'test_auc': None
        }

    def train(self, dataset, args, device: str = 'cuda'):

        self.trainer = DeepSADTrainer(device=device)
        # Get the model
        self.trainer.train(dataset, self.net, args)
        self.c = self.trainer.c

    def test(self, test_loader, delta_t, sigma, device):
        from utils import roc_auc_all
        import numpy as np
        self.net.eval()
        self.net.to(device)
        loss = []

        with torch.no_grad():
            for data in test_loader:
                
                x = data.to(device)
                x = torch.transpose(x, dim0=2, dim1=3)
                inputs = x.reshape(x.shape[0], x.shape[1]*x.shape[2], x.shape[3])
                outputs = self.net(inputs).squeeze().mean(dim=-1)
                batch_loss= torch.sum((outputs - self.c) ** 2, dim=1).cpu().numpy()
                loss.append(batch_loss)
        loss = np.concatenate(loss)

        auc_score, fps,tps = roc_auc_all(loss, delta_t, sigma)
        print("mean: {:.4f}, median: {:.4f}, auc:{:.4f}".format(np.mean(loss), np.median(loss),auc_score))# %%
        self.results['test_auc'] = auc_score
        return auc_score,fps,tps

    def pretrain(self, train_loader, args, device):

        self.ae_trainer = AETrainer(device=device)
        self.ae_net = self.ae_trainer.train(train_loader, self.ae_net, args)
        self.net = self.ae_net.encoder

    def save_model(self, export_model):
        """Save Deep SVDD model to export_model."""

        net_dict = self.net.state_dict()

        torch.save({'c': self.c,
                    'net_dict': net_dict}, export_model)

    def load_model(self, model_path, load_ae=False):
        """Load Deep SVDD model from model_path."""

        model_dict = torch.load(model_path)

        self.c = model_dict['c']
        self.net.load_state_dict(model_dict['net_dict'])

    def save_results(self, export_json):
        """Save results dict to a JSON-file."""
        with open(export_json, 'w') as fp:
            json.dump(self.results, fp)
# %%



from sklearn import metrics
from mtsa.utils import (
    files_train_test_split
)
from mtsa import (
    FEATURES,
    FINAL_MODEL,
    MFCCMix, 
    MFCCMixCV
)


from sklearn.mixture import GaussianMixture
import glob
import os
from functools import reduce
import numpy as np

from sklearn.manifold import TSNE
class MFCCMixPipeline():

    def __init__(
        self, 
        # final_model=GaussianMixture,
        # features=FEATURES,
        # sampling_rate=16000,
        *args, 
        **kwargs) -> None:
        self.__dict__.update(kwargs)
        # self.final_model = FINAL_MODEL
        
        final_model = GaussianMixture(n_components=self.n_components)
        self.mtsa = MFCCMix(
            sampling_rate=self.sampling_rate,   
            features=self.features,
            final_model=final_model
        )

        # self.atelier = AtelierCV(
        #     final_model=final_model,
        #     sampling_rate=self.sampling_rate   
        # )

    def calculate_aucroc(self, X_train, X_test, y_train, y_test):
        self.mtsa.fit(X_train, y_train)
        fpr, tpr, thresholds = metrics.roc_curve(y_test, self.tsa.score_samples(X_test))
        auc = metrics.auc(fpr, tpr)
        return auc
    
    
    def plot_tsne(self, X_train, X_test, y_train, y_test, **kwargs):
        self.mtsa.fit(X_train, y_train)
        
        perplexity = 4
        if 'perplexity' in kwargs:
            perplexity = kwargs['perplexity']
        tsne = TSNE(n_components=2, verbose=1, perplexity=perplexity, n_iter=300, random_state=None)
        
        X = self.mtsa.model['features'].fit_transform(
            self.mtsa.model['array2mfcc'].fit_transform(
            self.mtsa.model['wav2array'].fit_transform(X_test)))
        
        tsne_results = tsne.fit_transform(X)
        x = tsne_results[:,0]
        y = tsne_results[:,1]
        hue = y_test
        return x, y, hue
        

    def plot_tsne_individual(self, filepath, **kwargs):
        X_train, X_test, y_train, y_test = files_train_test_split(filepath)
        x, y, hue = self.plot_tsne(X_train, X_test, y_train, y_test, **kwargs)
        return x, y, hue


    # def calculate_aucroc_individual(self, filepath):
    #     X_train, X_test, y_train, y_test = files_train_test_split(filepath)
    #     roc = self.calculate_aucroc(X_train, X_test, y_train, y_test)
    #     return roc

    def calculate_aucroc_combined(self, filepath):
        def reduce_data(d1, d2):
            data = zip(d1, d2)
            data = list(data)
            X_train, X_test, y_train, y_test = [np.concatenate([d1, d2]) for (d1,d2) in data]
            return X_train, X_test, y_train, y_test
        paths = glob.glob(os.path.join(filepath, f'*{os.sep}' ))
        data = list(map(files_train_test_split, paths))
        X_train, X_test, y_train, y_test = reduce(reduce_data, data)
        roc = self.calculate_aucroc(X_train, X_test, y_train, y_test)
        return roc
   


from tensorflow.python.keras.layers import Dense, Layer

class Decoder(Layer):
    def __init__(self, output_dim: int, activation: str, output_activation: str,depth: int, **kwargs):
        super(Decoder, self).__init__(**kwargs)
        self.output_dim = output_dim
        self.activation = activation
        self.output_activation = output_activation
        self.depth = depth
        self.kwargs = kwargs
        
    def build(self, input_shape):
        self.hidden = {
            'hidden_{}'.format(i): Dense(
                int(self.output_dim/ (2**(self.depth-i))), activation=self.activation,
            ) for i in range(self.depth)
        }
        self.restored = Dense(self.output_dim, activation=self.output_activation)
        
    def call(self, inputs):
        x = self.hidden['hidden_0'](inputs)
        for i in range(1, self.depth):
            x = self.hidden['hidden_{}'.format(i)](x)
        return self.restored(x)
    
    def get_config(self):
        base_config = super(Decoder, self).get_config()
        config = {
            "output_dim": self.output_dim, 
            "activation": self.activation, 
            "output_activation": self.output_activation, 
            "depth": self.depth,
        }
        return dict(list(base_config.items()) + list(config.items()))



from keras.models import Model
from keras.layers import Input, Dense
from sklearn.base import BaseEstimator, OutlierMixin
import numpy as np
from mtsa.utils import Demux2Array, Wav2Array
from mtsa.features.mel import Array2MelSpec
from sklearn.pipeline import Pipeline
from functools import reduce
from keras.optimizers import Adam
import time

class AutoEncoderMixin(Model):
    def score_samples(self, X):
        return -1 * np.mean(np.square(X - self.predict(X))) 
    
    def fit(self, x=None, y=None, batch_size=None, epochs=50, verbose=0, callbacks=None, validation_split=0, validation_data=None, shuffle=True, class_weight=None, sample_weight=None, initial_epoch=0, steps_per_epoch=None, validation_steps=None, validation_batch_size=None, validation_freq=1, max_queue_size=10, workers=1, use_multiprocessing=True):
        #TODO final_model__epochs
        return super().fit(x=x, y=x, batch_size=batch_size, epochs=epochs, verbose=verbose, callbacks=callbacks, validation_split=validation_split, validation_data=validation_data, shuffle=shuffle, class_weight=class_weight, sample_weight=sample_weight, initial_epoch=initial_epoch, steps_per_epoch=steps_per_epoch, validation_steps=validation_steps, validation_batch_size=validation_batch_size, validation_freq=validation_freq)
    
       
class Hitachi(BaseEstimator, OutlierMixin):
    """
    
      :References:
    Purohit, Harsh, et al. "MIMII Dataset: Sound dataset for malfunctioning industrial machine investigation and inspection." arXiv preprint arXiv:1909.09347 (2019).
    
    """

    def __init__(self, 
                 sampling_rate=None,
                 random_state = None,
                 n_mels=64,
                 frames=5,
                 n_fft=1024,
                 hop_length=512,
                 power=2.0,
                 mono=False,
                 epochs=50,
                 batch_size=512,
                 learning_rate=1e-3, 
                 shuffle=True,
                 validation_split=0.1,
                 verbose=0
                 ) -> None:
        self.sampling_rate = sampling_rate
        self.random_state = random_state
        self.n_mels=n_mels
        self.frames=frames
        self.n_fft=n_fft
        self.hop_length=hop_length
        self.power=power
        self.mono = mono
        self.epochs=epochs
        self.batch_size=batch_size
        self.learning_rate = learning_rate
        self.shuffle=shuffle
        self.validation_split=validation_split
        self.verbose=verbose
        self.last_fit_time = 0
        self.model = self._build_model()
    

    @property
    def name(self):
        return "Hitachi"

    def fit(self, X, y=None):
        start = time.perf_counter()    
        
        self.model.fit(X, 
                              y,
                              final_model__batch_size=self.batch_size,
                              final_model__shuffle= self.shuffle,
                              final_model__validation_split=self.validation_split,
                              final_model__epochs = self.epochs,
                              final_model__verbose = self.verbose
                              )
        end = time.perf_counter()
        self.last_fit_time = end - start
        return self.last_fit_time #seconds
        
        

    def transform(self, X, y=None):

        def transform_one(Xi):
            l = list()
            l.append(Xi)
            l.extend(self.model.steps[:-1])
            Xt = reduce(lambda x, y: y[1].transform(x), l)
            return Xt



        Xp = np.array(
            list(
                map(
                    self.predict, 
                    [[x] for x in X])
                )
            )


        Xt = np.array(
            list(
                map(
                    transform_one, 
                    [[x] for x in X])
                )
            )

        Xf = np.mean(Xt - Xp, axis=1)
        return Xf

    def predict(self, X):
        return self.model.predict(X, verbose=0)

    def score_samples(self, X):
        return np.array(
            list(
                map(
                    self.model.score_samples, 
                    [[x] for x in X])
                )
            )
  
    
    def get_model(self):
        inputDim = self.n_mels * self.frames
        inputLayer = Input(shape=(inputDim,))
        h = Dense(64, activation="relu")(inputLayer)
        h = Dense(64, activation="relu")(h)
        h = Dense(8, activation="relu")(h)
        h = Dense(64, activation="relu")(h)
        h = Dense(64, activation="relu")(h)
        h = Dense(inputDim, activation=None)(h)
        final_model = AutoEncoderMixin(inputs=inputLayer, outputs=h)
        optimizer = Adam(learning_rate=self.learning_rate)
        final_model.compile(optimizer=optimizer, loss='mean_squared_error', jit_compile=True)
        return final_model
    
    def _build_model(self):
        wav2array = Wav2Array(
            sampling_rate=self.sampling_rate,
            mono=self.mono,
            )
        
        demux2array = Demux2Array()
        array2melspec= Array2MelSpec(
            sampling_rate=self.sampling_rate,
            n_fft=self.n_fft,
            hop_length=self.hop_length,
            n_mels=self.n_mels,
            frames=self.frames,
            power=self.power,
            )
        
        final_model = self.get_model()
        
        model = Pipeline(
            steps=[
                ("wav2array", wav2array),
                ("demux2array", demux2array),
                ("array2melspec", array2melspec),
                ("final_model", final_model),
                ]
            )
        
        return model


class HitachiDCASE2020(Hitachi):
    

    def get_model(self):
        inputDim = self.n_mels * self.frames
        inputLayer = Input(shape=(inputDim,))
        h = Dense(128, activation="relu")(inputLayer)
        h = Dense(128, activation="relu")(h)
        h = Dense(128, activation="relu")(h)
        h = Dense(128, activation="relu")(h)
        h = Dense(8, activation="relu")(h)
        h = Dense(128, activation="relu")(inputLayer)
        h = Dense(128, activation="relu")(h)
        h = Dense(128, activation="relu")(h)
        h = Dense(128, activation="relu")(h)
        h = Dense(inputDim, activation="relu")(h)
        final_model = AutoEncoderMixin(inputs=inputLayer, outputs=h)
        optimizer = Adam(learning_rate=self.learning_rate)  
        final_model.compile(optimizer=optimizer, loss='mean_squared_error')
        return final_model
    

def get_models_hitachi():
    hitachi = [("Baseline", Hitachi())]
    return hitachi


import numpy as np
import pandas as pd
import time
from sklearn.base import BaseEstimator, OutlierMixin, check_array
from sklearn.pipeline import (
    Pipeline, 
    FeatureUnion
) 
from mtsa.features.stats import (
    FEATURES,
    get_features
    )

from mtsa.features.mel import (
    Array2Mfcc 
)
from mtsa.utils import (
    Wav2Array,
)

from sklearn.ensemble import IsolationForest
from functools import reduce

class IForest(BaseEstimator, OutlierMixin):
    """
    Parâmetros:
        n_estimators (int): Número de árvores de isolamento a serem construídas.
        max_samples (int, float, str): Número de amostras a serem usadas para construir cada árvore.
        contamination (float, str): Estimativa de outliers no conjunto de dados.
        max_features (float, int): Número máximo de características a serem consideradas para cada divisão em uma árvore.
        bootstrap (bool): Se True, amostras são extraídas com substituição. Se false, cada árvore usa amostras únicas.
        n_jobs (int): Número de jobs a serem executados em paralelo.
        random_state (int, RandomState instance, None): Semente para gerador de números aleatórios.
        verbose (int): Controle da verbosidade.
        warm_start (bool): Reutiliza a solução do ajuste anterior para adicionar mais estimadores.
        final_model (estimator): Modelo final.
        features (list): Lista de características.
        sampling_rate (int): Taxa de amostragem para processamento de áudio.
    """

    def __init__(self,
                 n_estimators=100,
                 max_samples=256,
                 contamination=0.01,
                 max_features=1.0,
                 bootstrap=False,
                 n_jobs=None,
                 random_state=None,
                 verbose=0,
                 warm_start=False, 
                 features=FEATURES,
                 sampling_rate=None, 
                 ) -> None:
        super().__init__()
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.contamination = contamination
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.sampling_rate = sampling_rate
        self.features = features
        self.last_fit_time = 0
        self.execution_duration = None
        self.model_parameters_names = None
        self.dataset_name = None
        self.experiment_dataframe = self.__get_initialize_dataframe_experiments_result()
        self.model = self._build_model()

    @property
    def name(self):
        return "IsolationForest " + "+".join([f[0] for f in self.features])
        
    def fit(self, X, y=None):        
        start = time.perf_counter()    
        self.model.fit(X, y)
        end = time.perf_counter()
        self.last_fit_time = end - start
        return self.last_fit_time #seconds

    def transform(self, X, y=None):
        l = list()
        l.append(X)
        l.extend(self.model.steps[:-1])
        Xt = reduce(lambda x, y: y[1].transform(x), l)
        return Xt
    
    def predict(self, X):
        _predict = self.model.predict(X)
        _predict = np.where(_predict == -1, 0, 1)
        return _predict
    
    #def decision_function(self, X):
        """
        Retorna a pontuação de decisão para cada ponto de dados.
        
        Parâmetros:
            - X (array-like): Dados de entrada.
        
        Retorna:
            array-like: Pontuações de decisão.
        """
    #  return self.model.decision_function(X)
    
    def score(self, X, y=None):
        return self.model.score(X)

    # Retorna a pontuação de anomalia para cada ponto de dados
    def score_samples(self, X):
        return self.model.score_samples(X=X)
    
    def __get_initialize_dataframe_experiments_result(self):
        """
        Inicializa o DataFrame para armazenar os resultados dos experimentos.
        
        Retorna:
            pd.DataFrame: DataFrame vazio com as colunas definidas.
        """
        parameters_columns = self.__get_parameters_columns()                              
        return pd.DataFrame(columns=parameters_columns)
    
    def __get_parameters_columns(self):
        """
        Define as colunas do DataFrame de experimentos.
        
        Retorna:
            list: Lista de nomes de colunas.
        """
        parameters_columns = ["actual_dataset",
                              "parameters_names",
                              "n_estimators",
                              "max_samples",
                              "contamination",
                              "max_features",
                              "execution_duration",
                              "AUC_ROC",
                            ]
        return parameters_columns   

    def __create_dataframe(self):
        """
        Adiciona uma nova linha ao DataFrame de experimentos com os parâmetros e resultados atuais.
        """
        self.experiment_dataframe.loc[len(self.experiment_dataframe)] = {
            "actual_dataset": self.dataset_name,
            "parameters_names": self.model_parameters_names,
            "n_estimators": self.n_estimators, 
            "max_samples": self.max_samples, 
            "contamination": self.contamination, 
            "max_features": self.max_features,
            "execution_duration": self.execution_duration,
            "RMSE": None,
            "Score": None,
            "AUC_ROC": None
            } 
    
    def get_experiment_dataframe(self, dataset_name=None, model_parameters_names=None, execution_duration=None):
        self.dataset_name = dataset_name
        self.model_parameters_names = model_parameters_names
        self.__create_dataframe()
        return self.experiment_dataframe

    def _build_model(self):
        # Etapa 1: Converter áudio para array NumPy
        wav2array = Wav2Array(sampling_rate=self.sampling_rate)
        
        # Etapa 2: Converter array de áudio para MFCCs com n_mfcc especificado
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
        
        # Etapa 3: Combinar as features existentes usando FeatureUnion (Retorno virá 2D)
        features = FeatureUnion(self.features)

        # Configuração do IsolationForest
        self.final_model = IsolationForest(
            n_estimators=self.n_estimators, 
            max_samples=self.max_samples, 
            contamination=self.contamination,
            max_features=self.max_features,
            bootstrap=self.bootstrap,
            n_jobs=self.n_jobs, 
            random_state=self.random_state,
            verbose=self.verbose,
            warm_start=self.warm_start,
        )
        
        # Construção do pipeline
        model = Pipeline(
            steps=[
                ("wav2array", wav2array),
                ("array2mfcc", array2mfcc),
                ("features", features),
                ("final_model", self.final_model),
            ]
        )
        
        return model




from sklearn import metrics

def calculate_aucroc(model, X_test, y_test):
    y_hat=model.score_samples(X_test)
    fpr, tpr, thresholds = metrics.roc_curve(y_test, y_hat)
    auc = metrics.auc(fpr, tpr)
    return auc


import torch.nn as nn
import numpy as np
from tqdm import tqdm
import time

from functools import reduce

from sklearn import metrics
from sklearn.base import BaseEstimator, OutlierMixin
from sklearn.pipeline import Pipeline, FeatureUnion

from mtsa.features.mel import Array2Mfcc
from mtsa.models.GANF_components.ganfBaseModel import GANFBaseModel
from mtsa.utils import Wav2Array
from mtsa.features.stats import FEATURES


"""
    GRAPH-AUGMENTED NORMALIZING FLOWS

    This class implements the model described in the paper:
        GRAPH-AUGMENTED NORMALIZING FLOWS FOR ANOMALY DETECTION OF MULTIPLE TIME SERIES

    Key concepts and techniques from the paper have been adapted and implemented here.
    Any deviations or modifications from the original methodology are noted in the implementation details.

    Reference:
    [1] X. Wu, Y. Park, S. Tirthapura, "GRAPH-AUGMENTED NORMALIZING FLOWS FOR ANOMALY DETECTION OF MULTIPLE TIME SERIES,"
    AAAI, vol. 36, no. 8, 2022.  DOI: 10.1609/aaai.v36i8.20829
"""

class GANF(nn.Module, BaseEstimator, OutlierMixin):
    def __init__(
        self,
        sampling_rate=None,
        random_state=None,
        isForWaveData=False,
        batch_size=None,
        learning_rate=None,
        mono=True,
        use_array2mfcc=False,
        device=0,
        use_featureUnion= False
    ) -> None:
        super().__init__()
        self.sampling_rate = sampling_rate
        self.random_state = random_state
        self.isForWaveData = isForWaveData
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.mono = mono
        self.use_featureUnion = use_featureUnion
        self.final_model = GANFBaseModel(device=device)
        self.use_array2mfcc = use_array2mfcc
        self.model = self._build_model()
        self.last_fit_time = 0

    @property
    def name(self):
        return "GANF " + "+".join([f[0] for f in self.features])

    def fit(
        self,
        X,
        y=None,
        batch_size=None,
        epochs=None,
        max_iteraction=None,
        learning_rate=None,
        debug_dataframe=None,
    ):
        mono = self.mono and not self.use_array2mfcc
        if batch_size is None:
            batch_size = self.batch_size
        if learning_rate is None:
            learning_rate = self.learning_rate
        
        start = time.perf_counter()    
        self.model.fit(
            X,
            y,
            final_model__batch_size=batch_size,
            final_model__epochs=epochs,
            final_model__max_iteraction=max_iteraction,
            final_model__learning_rate=learning_rate,
            final_model__debug_dataframe=debug_dataframe,
            final_model__isWaveData=self.isForWaveData,
            final_model__mono=mono,
        )
        end = time.perf_counter()
        
        self.last_fit_time = end - start
        return self.last_fit_time #seconds

    def transform(self, X, y=None):
        l = list()
        l.append(X)
        l.extend(self.model.steps[:-1])
        Xt = reduce(lambda x, y: y[1].transform(x), l)
        return Xt

    def predict(self, X):
        return self.model.predict(X)

    def score_samples(self, X):
        return np.array(
            list(
                map(
                    self.model.score_samples, 
                    [[x] for x in X])
                )
            )

    def score(self, X, Y):
        X = self.score_samples
        fpr, tpr, thresholds = metrics.roc_curve(Y, X)
        auc = metrics.auc(fpr, tpr)
        return auc

    def get_adjacent_matrix(self):
        return self.final_model.get_adjacent_matrix()

    def _build_model(self):
        array2mfcc = Array2Mfcc(sampling_rate=self.sampling_rate)
        wav2array = Wav2Array(sampling_rate=self.sampling_rate, mono=self.mono)
        features = FeatureUnion(FEATURES)
        # mono = (self.mono and not self.use_array2mfcc)
        
        if self.use_featureUnion and self.use_array2mfcc and self.isForWaveData:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("array2mfcc", array2mfcc),
                    ("features", features),
                    ("final_model", self.final_model),
                ]
            )
        elif self.use_array2mfcc and self.isForWaveData:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("array2mfcc", array2mfcc),
                    ("final_model", self.final_model),
                ]
            )
        elif self.isForWaveData:
            model = Pipeline(
                steps=[
                    ("wav2array", wav2array),
                    ("final_model", self.final_model),
                ]
            )
        else:
            model = Pipeline(
                steps=[
                    ("final_model", self.final_model),
                ]
            )

        return model

    def attach_observer(self, observer):
        if (self.final_model is not None):
            self.final_model.attach_observer(observer)
            
    def set_adjacent_matrix(self, adjacent_matrix):
        if (self.final_model is not None):
            self.final_model.set_adjacent_matrix(adjacent_matrix)
            
    def get_random_adjacent_matrix(self):
         return self.final_model.get_random_adjacent_matrix()
     
    def get_initial_adjacent_matrix(self):
        return self.final_model.get_initial_adjacent_matrix()
    



from .oneClassSVM import OSVM
from .hitachi import *
from .mfccmix import *
from .ganf import *
from .ransyncorders import *
from .isolationforest import *
from .gacvae import *

__ALL__ = [
    Hitachi,
    HitachiDCASE2020,
    MFCCMix,
    GANF,
    RANSynCoders,
    IForest,
    GACVAE,
    OSVM
]




"""Base classes for all mfcc estimators."""

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.stats import kurtosis, skew
from scipy.fft import rfft
from functools import reduce
import itertools as ite
from mtsa.correlation_networks import PearsonCorrelationNetwork

#region feature extraction strategies

class MagnitudeMeanFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = X.mean(axis=2)
        return Xt
    
class MagnitudeStdFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = X.std(axis=2)
        return Xt
    
class CorrelationFeatureMfcc(BaseEstimator, TransformerMixin):
    
    def __init__(self) -> None:
        super().__init__()
    
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        def get_triu(G):
            return G[np.triu_indices(len(G), k = 1)]
        corr = PearsonCorrelationNetwork()
        Xt = map(corr.get_correlation, X)
        Xt = map(get_triu, Xt)
        Xt= np.array(list(Xt)) 
        return Xt
    
class RootMeanSquareFeature(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = np.sqrt(np.mean(X**2))
        return Xt
    
class SquareRootOfAmplitude(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = (np.mean(np.sqrt(np.absolute(X)))) ** 2
        return Xt
    
class Kurtosis(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = kurtosis(X)
        return Xt
    
class Skewness(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = skew(X)
        return Xt
    
class Peak2Peak(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = np.max(X) - np.min(X)
        return Xt
    
class CrestFactor(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.root_mean_square_feature = RootMeanSquareFeature()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = np.max(np.absolute(X)) / self.root_mean_square_feature.transform(X)
        return Xt
    
class ImpulseValue(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt =np.max(np.absolute(X)) / np.mean(np.absolute(X))
        return Xt
    
class MarginFactor(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.square_root_of_amplitude = SquareRootOfAmplitude()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = np.max(np.absolute(X)) / self.square_root_of_amplitude.transform(X)
        return Xt
    
class ShapeFactor(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.root_mean_square_feature = RootMeanSquareFeature()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = self.root_mean_square_feature.transform(X) / np.mean(np.absolute(X))
        return Xt

class KurtosisFactor(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.kurtosis = Kurtosis()
        self.root_mean_square_feature = RootMeanSquareFeature()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = self.kurtosis.transform(X) / (self.root_mean_square_feature.transform(X) ** 4)
        return Xt
    
class FrequencyCenter(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = 2 * np.abs(rfft(X)) / X.size
        return np.mean(Xt)
    
class RootMeanSquareFrequency(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = 2 * np.abs(rfft(X)) / X.size
        return np.sqrt(np.mean(Xt**2))
    
class RootFrequencyVariance(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.frequency_center = FrequencyCenter()
        
    def fit(self, X, y=None):
        return self
    
    def transform(self, X, y=None, **fit_params):
        Xt = 2 * np.abs(rfft(X)) / X.size
        return np.sqrt(np.mean((Xt - self.frequency_center.transform(Xt)) ** 2))

class FeatureExtractionMixer(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        super().__init__()
        self.strategies = []
    
    def fit(self, X, y=None):
        return self

    def append_strategy(self, feature_extraction_strategy):
        self.strategies.append(feature_extraction_strategy)
    
    def transform(self, X, y=None, **fit_params):
        samples = []
        for x in X:
            features = []
            for strategy in self.strategies: 
                features.append(float(strategy.transform(x)))
            features = np.array(features, dtype=float)
            samples.append(features)
            
        result = np.array(samples)
        return result
    
#endregion

FEATURES = [
    ("M", MagnitudeMeanFeatureMfcc()), 
    ("S", MagnitudeStdFeatureMfcc()), 
    ("C", CorrelationFeatureMfcc())
]

def get_features(self):
    number_features = np.arange(1, len(FEATURES)+1)
    def combinations(r): return ite.combinations(FEATURES, r)
    all_combinations = map(combinations, number_features)
    features = reduce(lambda x, y: ite.chain(x, y), all_combinations)
    return features




import numpy as np

class NetworkLearnerModel:
    def __init__(self):
        super().__init__()
        self.observers = []

    def attach_observer(self, observer):
        self.observers.append(observer)

    def dettach_observer(self, observer):
        self.observers.remove(observer)

    def notify_observers(self, adjacent_matrix, epoch, h, total_loss, adj_matrix_name, learning_rate, batch_size):
        self.__proxy_notify_observers(
            current_epoch=epoch,
            adjacent_matrix=adjacent_matrix.clone(),
            loss=total_loss.clone(),
            DAG_constraint_h=h.clone(),
            models_specs={
                "learning_rate:": learning_rate,
                "batch_size": batch_size,
            },
            AUC_ROC=-np.inf,
            adj_matrix_name=adj_matrix_name,
        )
        
    def __notifyNetworkExtractor(self, **kwargs):
        for observer in self.observers:
            observer.update(kwargs)

    def __proxy_notify_observers(self, **kwargs):
        schema_avro = {
            "type": "record",
            "name": "GANF_Network",
            "fields": [
                {"name": "adj_matrix_name", "type": "string"},
                {"name": "current_epoch", "type": "int"},
                {"name": "loss", "type": "float"},
                {"name": "DAG_constraint_h", "type": "float"},
                {
                    "name": "adjacent_matrix",
                    "type": {
                        "type": "array",
                        "items": {"type": "array", "items": "double"},
                    },
                },
                {"name": "models_specs", "type": {"type": "map", "values": "float"}},
                {"name": "AUC_ROC", "type": "float"},
            ],
        }
        data = [
            {
                "current_epoch": kwargs["current_epoch"],
                "loss": kwargs["loss"],
                "DAG_constraint_h": kwargs["DAG_constraint_h"],
                "adjacent_matrix": kwargs["adjacent_matrix"].tolist(),
                "models_specs": kwargs["models_specs"],
                "AUC_ROC": kwargs["AUC_ROC"],
                "adj_matrix_name": kwargs["adj_matrix_name"],
            }
        ]

        self.__notifyNetworkExtractor(schema_avro=schema_avro, data=data)



from abc import abstractmethod
import numpy as np
import matplotlib.pyplot as plt
from itertools import cycle
import itertools as it
import os
import abc
from sklearn import metrics

from sklearn.metrics import confusion_matrix
from sklearn import metrics
from sklearn.manifold import TSNE

def _get_par(name, default, **kwargs):
    if name in kwargs: return kwargs[name]
    return default

def get_tsne_results(model, X, y, **kwargs):
    n_components = _get_par('n_components', 2, **kwargs)
    perplexity = _get_par('perplexity', 4, **kwargs)
    n_iter = _get_par('n_iter', 300, **kwargs)
    random_state = _get_par('random_state', None, **kwargs)
    tsne = TSNE(n_components=n_components, verbose=0, perplexity=perplexity, n_iter=n_iter, random_state=random_state)
    
    # Xt = model['features'].fit_transform(
    #     model['array2mfcc'].fit_transform(
    #     model['wav2array'].fit_transform(X)))
    
    tsne_results = tsne.fit_transform(model.transform(X))
    p1 = tsne_results[:,0]
    p2 = tsne_results[:,1]
    p3 = y
    return p1, p2, p3
        

# def plot_tsne_individual(self, filepath, **kwargs):
#     X_train, X_test, y_train, y_test = files_train_test_split(filepath)
#     x, y, hue = self.plot_tsne(X_train, X_test, y_train, y_test, **kwargs)
#     return x, y, hue
    
    
    
class Plotter(abc.ABC):
    markers = ['D', 'o', 'x','s', '^', 'd', 'h', '+', '*', ',', '.', '1', 'p', '3', '2', '4', 'H', 'v', '8',
               '<', '>']
    # colors = ['g', 'y', 'r', 'b', 'k', 'm', 'c']\

    colors = ['#1f77b4', '#ff7f0e',  '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']


    def __init__(self, *args, **kwargs) -> None:
        self.__dict__.update(kwargs)
    
    def plot(self, **kwargs):
        if "path_output" in kwargs:
            path_output = kwargs["path_output"]
            dir_output = os.path.dirname(path_output)
            if not os.path.exists(dir_output):
                os.makedirs(dir_output)
        self.plot_elem(**kwargs)
    
    @abc.abstractmethod
    def plot_elem(self, **kwargs):
        """Implement """


   

   

    @staticmethod
    def plot_correlation(correlation_network, **kwargs):
        ticks = range(1, 21)
        plotter = PlotterHeatMap(X=correlation_network)
        plotter.plot(
            cmap="RdBu",
            xticks = np.arange(len(ticks)),
            yticks = np.arange(len(ticks)),
            xticklabels = ticks,
            yticklabels = ticks,
            xlabel = "MFCC Number",
            ylabel = "MFCC Number",
            # title = kwargs["title"],
            # path_output = kwargs["path_output"]
            **kwargs
        )

        
        # fig, ax = plt.subplots()
        # im = ax.imshow(correlation_network, cmap="RdBu")

        # ax.set_xticks(np.arange(len(x_ticks)))
        # ax.set_yticks(np.arange(len(x_ticks)))

        # ax.set_xticklabels(x_ticks)
        # ax.set_yticklabels(x_ticks)

        # ax.set_xlabel("MFCC Number")
        # ax.set_ylabel("MFCC Number")


        # if "title" in kwargs:
        #     ax.set_title(kwargs["title"])
        
        # cbar=plt.colorbar(im)
        # fig.tight_layout()

        # if "path_output" in kwargs:
        #     plt.savefig(kwargs["path_output"])
        # else: 
        #     plt.show()
        # plt.close()
        

    @staticmethod
    def plot_confidence_interval(confidence_intervals, **kwargs):
        fig, ax = plt.subplots(1,1)
        colors = cycle(['r', 'g', 'b', 'y'])
        for i, ci in enumerate(confidence_intervals):
            lower = ci[0]
            upper = ci[1]
            mean = (upper + lower) / 2
            color = next(colors)
            ax.plot(i + 1, mean, marker='o', markersize=10, color=color)
            ax.vlines(x=i + 1, ymin=lower, ymax=upper, linewidth=2, color=color)
            ax.hlines(y=lower, xmin=i + 1 - 0.1, xmax=i + 1 + 0.1, linewidth=2, color=color)
            ax.hlines(y=upper, xmin=i + 1 - 0.1, xmax=i + 1 + 0.1, linewidth=2, color=color)
        
        ax.set_xticks(np.arange(1, len(confidence_intervals) + 1))
        ax.tick_params(axis='both', which='major', labelsize=15)
        
        ax.set_ylim(-1,1)

        if "title" in kwargs:
            plt.title(kwargs["title"], size=15)

        if "xticklabels" in kwargs:
            ax.set_xticklabels(kwargs["xticklabels"])

        if "ylabel" in kwargs:
            ax.set_ylabel(kwargs["ylabel"], size=20)

        if "xlabel" in kwargs:
            ax.set_xlabel(kwargs["xlabel"], size=20)

        
        plt.tight_layout()
        if "path_output" in kwargs:
            plt.savefig(kwargs["path_output"])
        else: 
            plt.show()
        plt.close()

    # @abstractmethod
    # def plot_confidence_interval_matrix(confidence_intervals, **kwargs):
    #     fig, ax = plt.subplots(5,5)
    #     ijs = (ij for ij in it.product(range(20), range(20)) if ij[0] != ij[1])
    #     for (i, j) in ijs:
    #         colors = cycle(['r', 'g', 'b', 'y'])
    #         for k in range(len(confidence_intervals)):
    #             lower = confidence_intervals[k,0,i,j]
    #             upper = confidence_intervals[k,1,i,j]
    #             mean = (upper + lower) / 2
    #             color = next(colors)
    #             ax[i,j].plot(k + 1, mean, marker='o', color=color)
    #             ax[i,j].vlines(x=k + 1, ymin=lower, ymax=upper, color=color)
    #             ax[i,j].hlines(y=lower, xmin=k + 1 - 0.1, xmax=k + 1 + 0.1, linewidth=2, color=color)
    #             ax[i,j].hlines(y=upper, xmin=k + 1 - 0.1, xmax=k + 1 + 0.1, linewidth=2, color=color)
            
    #         ax[i,j].set_xticks(np.arange(1, len(confidence_intervals) + 1))
    #         ax[i,j].tick_params(axis='both', which='major')
    #         ax[i,j].set_ylim(-1,1)

    #     # if "title" in kwargs:
    #     #     plt.title(kwargs["title"], size=20)

    #     # if "labels" in kwargs:
    #     #     ax.set_xticklabels(kwargs["labels"])

    #     # ax.set_xlim(0,3)
        
    #     # ax.set_xlabel("Estimators", size=20)
    #     # ax.set_ylabel("X", size=20)
        
        
    #     if "path_output" in kwargs:
    #         plt.savefig(kwargs["path_output"])
    #     else: 
    #         plt.show()
    #     plt.close()

class PlotterActivation(Plotter):

    def plot_elem(self, **kwargs):
        ticks = range(1, 21)
        plotter_heatmap = PlotterHeatMap(X=self.X)
        plotter_heatmap.plot(
            cmap="Blues",
            xticks = np.arange(len(ticks)),
            yticks = np.arange(len(ticks)),
            xticklabels = ticks,
            yticklabels = ticks,
            xlabel = "MFCC Number",
            ylabel = "MFCC Number",
            # title = kwargs["title"],
            path_output = self.path_output,
            **kwargs
        )

class PlotterHeatMap(Plotter):

    def plot_elem(self, **kwargs):
        fig, ax = plt.subplots()

        cmap="RdBu"
        if "cmap" in kwargs:
            cmap = kwargs["cmap"]

        im = ax.imshow(self.X, cmap=cmap)

        if "xticks" in kwargs:
            ax.set_xticks(kwargs["xticks"])
        
        if "yticks" in kwargs:
            ax.set_yticks(kwargs["yticks"])

        if "xlabel" in kwargs:
            ax.set_xlabel(kwargs["xlabel"])

        if "ylabel" in kwargs:
            ax.set_ylabel(kwargs["ylabel"])

        if "xticklabels" in kwargs:
            ax.set_xticklabels(kwargs["xticklabels"])
        
        if "yticklabels" in kwargs:
            ax.set_yticklabels(kwargs["yticklabels"])

        if "title" in kwargs:
            ax.set_title(kwargs["title"])
        
        cbar=plt.colorbar(im)
        fig.tight_layout()

        if "path_output" in kwargs:
            plt.savefig(kwargs["path_output"])
        else: 
            plt.show()
        plt.close()

class PlotterROC(Plotter):

    def plot_elem(self, **kwargs):
        fpr, tpr, thresholds = metrics.roc_curve(self.y_true, self.y_pred)
        auc = metrics.auc(fpr, tpr)
        # self.plot_roc(fpr, tpr, auc, title=self.title, path_output=self.path_ouput_file)

    # def plot_roc(self, **kwargs):
        
        title = ""
        if "title" in kwargs:
            title = kwargs["title"]

        fig, ax = plt.subplots()
        ax.plot(fpr, tpr, color='deeppink', linewidth=4, label=f"ROC Curve (area = {round(auc,2) })")
        ax.plot([0,1], [0,1], "k--",linewidth=4, label='random classifier')
        ax.set_xlim([0.0, 1.0])
        ax.set_ylim([0.0, 1.05])
        ax.set_xlabel("False Positive Rate (1 - Specificity)", fontsize=16)
        ax.set_ylabel("True Positive Rate (Sensitivity)", fontsize=16)
        ax.set_title(f"Receiver Operating Characteristic \n {title}", fontsize=18)
        ax.legend(loc="lower right")
        if "path_output" in kwargs:
            plt.savefig(kwargs["path_output"])
        plt.close()
